# Generated by Django 4.2.4 on 2023-08-03 14:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0006_tslmodel_entrypoint'),
    ]

    operations = [
        migrations.AddField(
            model_name='ocrmodel',
            name='entrypoint',
            field=models.CharField(max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='ocrmodel',
            name='default_options',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='used_by_%(class)s', to='ocr_translate.optiondict'),
        ),
        migrations.AlterField(
            model_name='tslmodel',
            name='default_options',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='used_by_%(class)s', to='ocr_translate.optiondict'),
        ),
    ]
